import gleam/dynamic.{type Decoder}
import gleam/http/request.{type Request}
import gleam/http/response.{type Response}
import gleam/httpc
import gleam/json
import gleam/pair
import gleam/result
import gleam/string

import apkallone.{type Client, type Error}

pub fn prepare_request(
  client: Client,
  endpoint: String,
) -> Result(Request(String), Error) {
  use req <- result.try(
    request.to(client.instance <> endpoint)
    |> result.replace_error(apkallone.BadInstance),
  )

  req
  |> request.prepend_header("user-agent", apkallone.user_agent)
  |> Ok
}

pub fn send_request(req: Request(String)) -> Result(Response(String), Error) {
  use res <- result.try(
    httpc.send(req)
    |> result.map_error(apkallone.FetchError),
  )

  case res.status {
    200 -> Ok(res)
    other -> Error(apkallone.UnexpectedStatus(other))
  }
}

pub fn decode_response(
  res: Response(String),
  decoder: Decoder(a),
) -> Result(a, Error) {
  use <- require_content_type(res, "application/json")

  res.body
  |> json.decode(decoder)
  |> result.map_error(apkallone.DecodeError)
}

fn require_content_type(
  res: Response(String),
  expected_type: String,
  fun: fn() -> Result(a, Error),
) -> Result(a, Error) {
  use actual_type <- result.try(
    response.get_header(res, "content-type")
    |> result.replace_error(apkallone.MissingContentType),
  )

  let actual_type =
    string.split_once(actual_type, on: ";")
    |> result.map(pair.first)
    |> result.unwrap(or: actual_type)

  case actual_type == expected_type {
    True -> fun()
    False -> Error(apkallone.UnexpectedContentType(actual_type))
  }
}
